<?php

namespace App\Controller;

use AltoRouter;
use App\Http\HttpCodes;
use App\Http\HttpResponse;
use App\Http\JsonHttpResponse;
use App\Http\ViewHttpResponse;
use App\Session\MutableSessionHandle;
use App\Validation\ValidationFail;
use Exception;
use Twig\Environment;
use Twig\Error\LoaderError;
use Twig\Error\RuntimeError;
use Twig\Error\SyntaxError;
use Twig\Loader\FilesystemLoader;

class FrontController {
    private AltoRouter $router;
    private string $basePath;

    private const USER_CONTROLLER = "UserController";
    private const VISITOR_CONTROLLER = "VisitorController";


    public function __construct(string $basePath) {
        $this->router = $this->createRouter($basePath);
        $this->initializeRouterMap();
        $this->basePath = $basePath;
    }

    /**
     * @param MutableSessionHandle $session
     * @return void
     * @throws LoaderError
     * @throws RuntimeError
     * @throws SyntaxError
     */
    public function run(MutableSessionHandle $session): void {
        $match = $this->router->match();
        if ($match) {
            $this->handleMatch($match, $session);
            return;
        }

        $this->displayViewByKind(ViewHttpResponse::twig("error.html.twig", [
            'failures' => [ValidationFail::notFound("Could not find page ${_SERVER['REQUEST_URI']}.")],
        ], HttpCodes::NOT_FOUND));
    }

    /**
     * Create a new instance of an AltoRouter
     *
     * @param string $basePath
     * @return AltoRouter
     */
    public function createRouter(string $basePath): AltoRouter {
        $router = new AltoRouter();
        $router->setBasePath($basePath);
        return $router;
    }

    /**
     * Initialize project's routes
     *
     * @return void
     */
    private function initializeRouterMap(): void {
        $this->router->map("GET", "/home", self::USER_CONTROLLER);
        $this->router->map("GET|POST", "/user/[a:action]/[i:idTactic]?", self::USER_CONTROLLER);
        $this->router->map("GET|POST", "/visitor/[a:action]", self::VISITOR_CONTROLLER);
    }

    /**
     * @param array<string, mixed> $match
     * @param MutableSessionHandle $session
     * @return void
     */
    private function handleMatch(array $match, MutableSessionHandle $session): void {
        $tag = $match['target'];

        $action = $this->getAction($match);
        $params = $match["params"];
        unset($params['action']);
        $this->handleResponseByType($this->tryToCall($tag, $action, array_values($params), $session));
    }


    /**
     * @param string $controllerName
     * @param string $action
     * @param array<int, mixed> $params
     * @param MutableSessionHandle $session
     * @return HttpResponse
     */
    private function tryToCall(string $controllerName, string $action, array $params, MutableSessionHandle $session): HttpResponse {
        if ($controllerName != self::VISITOR_CONTROLLER) {
            $account = $session->getAccount();
            if ($account == null) {
                // put in the session the initial url the user wanted to get
                $session->setInitialTarget($_SERVER['REQUEST_URI']);
                return HttpResponse::redirect($this->basePath . "/visitor/login");
            }
        }

        $controller = $this->getController($controllerName);

        if (is_callable([$controller, $action])) {
            // append the session as the last parameter of a controller function
            $params[] = $session;
            return call_user_func_array([$controller, $action], $params);
        } else {
            return ViewHttpResponse::twig("error.html.twig", [
                'failures' => [ValidationFail::notFound("Could not find page ${_SERVER['REQUEST_URI']}.")],
            ], HttpCodes::NOT_FOUND);
        }
    }

    /**
     * Get the right method to call to do an action
     *
     * @param array<string, mixed> $match
     * @return string
     */
    private function getAction(array $match): string {
        if (isset($match["params"]["action"])) {
            return $match["params"]["action"];
        }
        return "home";
    }

    /**
     * Initialize the right controller by the user's role
     *
     * @param string $controller
     * @return mixed
     */
    private function getController(string $controller) {
        $namespace = "\\App\\Controller\\";
        $controller = $namespace . $controller;
        return new $controller();
    }

    /**
     * Redirect the return by the response's type
     *
     * @param HttpResponse $response
     * @return void
     */
    private function handleResponseByType(HttpResponse $response): void {
        http_response_code($response->getCode());

        foreach ($response->getHeaders() as $header => $value) {
            header("$header: $value");
        }

        if ($response instanceof ViewHttpResponse) {
            $this->displayViewByKind($response);
        } elseif ($response instanceof JsonHttpResponse) {
            header('Content-type: application/json');
            echo $response->getJson();
        }
    }

    /**
     * Use the right method to display the response
     *
     * @param ViewHttpResponse $response
     * @return void
     */
    private function displayViewByKind(ViewHttpResponse $response): void {
        $file = $response->getFile();
        $args = $response->getArguments();

        switch ($response->getViewKind()) {
            case ViewHttpResponse::REACT_VIEW:
                send_react_front($file, $args);
                break;
            case ViewHttpResponse::TWIG_VIEW:
                try {
                    $loader = new FilesystemLoader('../src/Views/');
                    $twig = new Environment($loader);
                    $twig->display($file, $args);
                } catch (RuntimeError | SyntaxError | LoaderError $e) {
                    http_response_code(500);
                    echo "There was an error rendering your view, please refer to an administrator.\nlogs date: " . date("YYYD, d M Y H:i:s");
                    throw $e;
                }
                break;
        }
    }
}
